 /*******************************************************************************
*  transfer.c            Interface procedures for AGATA digitizer communications
*/


#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <winsock2.h>
#endif

#ifdef UNIX
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <ctype.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>

#include <errno.h>

#include "transfer.h"

#define VERSION "0.3"

#define MAXUSERS 32
int CurrentUser = 0;

#define XPORT_TCP_PORT 10001

int SendTimeout = 10000;      /*  default is 10 seconds  */
int RecvTimeout = 10000;      /*  default is 10 seconds  */

unsigned int UserSendTimeout[MAXUSERS];
unsigned int UserRecvTimeout[MAXUSERS];

unsigned short Port = 0;       /* not defined initially */
unsigned short UserPort[MAXUSERS];

char XportHost[80];
unsigned char UserXportHost[MAXUSERS][80];

char MsgBuf[256];

int verbose = 0;

int BlockSize = 1024;         /*  units bytes  -  1K bytes is default */
int UserBlockSize[MAXUSERS];

#define MINSNDBUFSIZE 32          /*  units Kbytes  */
#define MAXSNDBUFSIZE 128      /* default max attempted */
#define MINRCVBUFSIZE 32          /*  units Kbytes  */
#define MAXRCVBUFSIZE 128      /* default max attempted */

int SndBufSize = MAXSNDBUFSIZE;   /*  units Kbytes */
int RcvBufSize = MAXRCVBUFSIZE;   /*  units Kbytes */

#ifdef WIN32
    WSADATA wsaData;
#else
    #define SOCKET int
    #define SOCKET_ERROR -1
#endif

#define TIMEOUT -2
#define CLOSED -3
#define OK 0
#define XPORT_ERROR 1

struct sockaddr_in  server;

SOCKET tsSocket = 0;
SOCKET UsertsSocket[MAXUSERS];

int clientInit (char *);

#ifdef WIN32
BOOL WINAPI DllMain (HANDLE hModule, DWORD dwFunction, LPVOID lpNot) {

    return TRUE;
}
#endif

#ifdef WIN32
void ReportError () {

  LPVOID lpMsgBuf;
  DWORD ErrNo;
  
       ErrNo = WSAGetLastError();
       FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
                      NULL,
                      ErrNo,
                      0,
                      (LPTSTR) &lpMsgBuf,
                      0,
                      NULL
                      );
                      
       sprintf(MsgBuf, "Transfer Error %d - %s", ErrNo, lpMsgBuf);
       fprintf(stdout, "%s", MsgBuf);
       LocalFree (lpMsgBuf);
}
#else
void ReportError () {
       perror("Transfer Error - ");
       sprintf(MsgBuf, "Error %d", errno);
}
#endif

/*******************************************************************************/

void dump(char *p, int nb)
{
int i, j, k;
unsigned short *pShort;
unsigned char *pChar;  

   j = nb / 16;
   if ( j*16 < nb ) j++;

   pShort = (unsigned short *)p;
   pChar = (unsigned char *)p;
   
   for (i = 0; i < j; i++)
   {

      printf("\n%8x:", pChar);
      for (k = 0; k < 8; k++)
         printf(" %04x", *pShort++);
         
      printf("    ");
      
      for (k = 0; k < 16; k++)
      {
         if ((*pChar >= 32) && (*pChar < 127))
            printf("%c", *pChar);
         else
             printf(".");
             
         pChar++;
         
      }  

   }
   printf("\n\r");  
} 

/*******************************************************************************

    same parameter list as send
    loops until all data has been sent  (or ERROR)
*/
int do_send(SOCKET S, char * p, int l)
{
   int rc, done, left;
   fd_set ready;
   struct timeval timeout;

    done = 0;
    left =  l;

    while (done < l) {

        FD_ZERO(&ready);
        FD_SET(S,&ready);

        timeout.tv_sec = SendTimeout/1000;    /* apply a  timeout */
        timeout.tv_usec = SendTimeout - ((SendTimeout/1000)*1000);

        rc = select(S+1, (fd_set *)0, &ready, (fd_set *)0, &timeout);
        if (rc < 0) {
           ReportError();
           sprintf(MsgBuf, "select() failed:  ");
           fprintf (stdout, "%s\n", MsgBuf);
           return(SOCKET_ERROR);
        }
        if (rc == 0) {
           sprintf(MsgBuf, "select() returned no ready sockets");
           fprintf (stdout, "%s\n", MsgBuf);
           return(TIMEOUT);
        }
        if (!(FD_ISSET(S, &ready))) return(SOCKET_ERROR);

        rc = send(S, p+done, left, 0);
        if (rc == SOCKET_ERROR)
             return rc;
        done += rc;
        left -= rc;
    }

    return OK;
}

/*******************************************************************************
    same parameter list as recv
    loops until all data has been received  (or ERROR)
*/
int do_recv(SOCKET S, char * p, int * l, int t)
{
   int rc, done, left, length;
   fd_set ready;
   struct timeval timeout;

    length = *l;

    done = 0;
    left =  length;

    *l = done;

    while (done < length) {

        FD_ZERO(&ready);
        FD_SET(S,&ready);

        timeout.tv_sec = RecvTimeout/1000;    /* apply a  timeout */
        timeout.tv_usec = RecvTimeout - ((RecvTimeout/1000)*1000);

        rc = select(S+1, &ready, (fd_set *)0, (fd_set *)0, &timeout);
        if (rc < 0) {
           ReportError();
           sprintf(MsgBuf, "select() failed:  ");
           fprintf (stdout, "%s\n", MsgBuf);
           return(SOCKET_ERROR);
        }
        if (rc == 0) {
           sprintf(MsgBuf, "select() returned no ready sockets");
           fprintf (stdout, "%s\n", MsgBuf);
           return(TIMEOUT);
        }
        if (!(FD_ISSET(S, &ready))) return(SOCKET_ERROR);

        rc = recv(S, p+done, left, t);
        if (rc == SOCKET_ERROR) return SOCKET_ERROR;
        if (rc == 0) return CLOSED;
        done += rc;
        left -= rc;
        *l = done;
    }

    return OK;
}
/*******************************************************************************
* XportResetAll   initialise software
*/
void XportResetAll ()
{
    int i;

    XportReset ();

    for (i = 0; i < MAXUSERS; i++)
    {
        UserPort[i] = Port;
        UserBlockSize[i] = BlockSize;
        UsertsSocket[i] = tsSocket;
        UserSendTimeout[i] = SendTimeout;
        UserRecvTimeout[i] = RecvTimeout;
    }
    return;
}/*******************************************************************************
* XportReset   initialise current user
*/
void XportReset ()
{
    Port = XPORT_TCP_PORT;
    BlockSize = 1024;
    tsSocket = 0;
    SendTimeout = 5000;
    RecvTimeout = 5000;

    return;
}/*******************************************************************************
* XportInit   initialise interface
*            return OK or ERROR
*    IPStr is a char* to the name of the Xport device
*/
int XportInit (char *IPStr)
{
    int rs;
    
    strcpy (XportHost, IPStr);

    rs = clientInit (XportHost);

    return rs;
}/*******************************************************************************
* XportSendTimeout        allows a custom value to be set
*/
int XportSendTimeout(int n)
{
    SendTimeout = n;
    if (verbose) printf("Setting Send Timeout %d millisecs\n",SendTimeout);
    return OK;
}
/*******************************************************************************
* XportREcvTimeout        allows a custom value to be set
*/
int XportRecvTimeout(int n)
{
    RecvTimeout = n;
    if (verbose) printf("Setting Recv Timeout %d millisecs\n",RecvTimeout);
    return OK;
}
/*******************************************************************************
* XportBlockSize        allows a custom size to be set
*/
int XportBlockSize(int n)
{
    BlockSize = n;
    if (verbose) printf("Setting Transfer Block Size %d\n",BlockSize);
    return OK;
}
/*******************************************************************************
* XportSndBufSize        allows a custom size to be set
*/
int XportSndBufSize(int n)
{
    SndBufSize = n;
    if (verbose) printf("Setting TCP send buffer size %dK\n",SndBufSize);
    return OK;
}
/*******************************************************************************
* XportRcvBufSize        allows a custom size to be set
*/
int XportRcvBufSize(int n)
{
    RcvBufSize = n;
    if (verbose) printf("Setting TCP receive buffer size %dK\n",RcvBufSize);
    return OK;
}
/*******************************************************************************
* XportPort        allows a custom port to be set
*/
int XportPort(int n)
{
    Port = (unsigned short) n;
    return OK;
}
/*******************************************************************************
* XportSetVerbose        allows more debug information 
*/
int XportSetVerbose(int n)
{
    verbose = (unsigned short) n;
    return OK;
}
/*******************************************************************************
* XportSetUser        allows current user to be set 
*/
int XportSetUser(int n)
{
    int i;

    if (n>MAXUSERS) return XPORT_ERROR;
    if (n == CurrentUser) return OK;

/*    change user  */

/*    save current user  */
    UserPort[CurrentUser] = Port;
    UserBlockSize[CurrentUser] = BlockSize;
    UsertsSocket[CurrentUser] = tsSocket;
    UserSendTimeout[CurrentUser] = SendTimeout;
    UserRecvTimeout[CurrentUser] = RecvTimeout;

    for (i = 0; i < 80; i++)
    {
        UserXportHost[CurrentUser][i] = XportHost[i];
    }

/*    load new user  */
    CurrentUser = (unsigned short) n;
    Port = UserPort[CurrentUser];
    BlockSize = UserBlockSize[CurrentUser];
    tsSocket = UsertsSocket[CurrentUser];
    SendTimeout = UserSendTimeout[CurrentUser];
    RecvTimeout = UserRecvTimeout[CurrentUser];

    for (i = 0; i < 80; i++)
    {
        XportHost[i] = UserXportHost[CurrentUser][i];
    }

    return OK;
}
/*******************************************************************************
* XportOpenClient            initial function to establish host/client connection
*                        create a socket
*                        find the host
*                        create a client
*                        returns client pointer or NULL
*/
int XportOpenClient (char *IPStr)
{         
       int retval;
       struct hostent *hp;
       unsigned int addr;
       struct sockaddr_in server;
       int sndsize, sndreq, rcvsize, rcvreq, temp, length;

#ifdef WIN32
    int yes=1;
#endif

    sprintf(MsgBuf, "Xport Access version %s",VERSION);
    if (verbose) printf ("%s\n", MsgBuf);

#ifdef WIN32
    if ((retval = WSAStartup(MAKEWORD( 2, 2),&wsaData)) != 0) {
        WSASetLastError(retval);
        ReportError();
        sprintf(MsgBuf, "WSAStartup failed:");
        fprintf (stdout, "%s\n", MsgBuf);
        XportClose();
        return XPORT_ERROR;
    }
#endif

     if (Port == 0) Port = XPORT_TCP_PORT;

     memset(&server,0,sizeof(server));
     server.sin_port = htons(Port);
    
    /* Attempt to detect if we should call gethostbyname() or gethostbyaddr() */
    
    if (isalpha(IPStr[0])) {   /* server address is a name */
        hp = gethostbyname(IPStr);
        if (hp == NULL ) {
            sprintf(MsgBuf,"Cannot resolve address [%s]: ", IPStr);
            fprintf (stdout, "%s\n", MsgBuf);
            XportClose();
            return XPORT_ERROR;
        }
        server.sin_family = hp->h_addrtype;
        memcpy(&(server.sin_addr),hp->h_addr,hp->h_length);
    }
    else  { /* Convert nnn.nnn address to a usable one */
        addr = inet_addr(IPStr);
        server.sin_family = AF_INET;
        server.sin_addr.s_addr = addr;
    }

    tsSocket = socket(AF_INET, SOCK_STREAM, 0); /* Open a TCP socket */
    if (tsSocket < 0 ) {
        ReportError();
        sprintf(MsgBuf, "Client: Error Opening socket: ");
        fprintf (stdout, "%s\n", MsgBuf);
        XportClose();
        return XPORT_ERROR;
    }
    
/*    optimize TCP send buffer size    */

    length=sizeof(sndsize);
    if (getsockopt(tsSocket,SOL_SOCKET,SO_SNDBUF,(char *)&sndsize,&length) == SOCKET_ERROR) {
        ReportError();
        sprintf(MsgBuf, "getsockopt() failed:  ");
        fprintf (stdout, "%s\n", MsgBuf);
    }
    temp = sndsize;

    for (sndreq = MINSNDBUFSIZE; sndreq <= SndBufSize; sndreq = sndreq+MINSNDBUFSIZE) {
        sndsize = sndreq*1024;
        if (setsockopt(tsSocket,SOL_SOCKET,SO_SNDBUF,(char *)&sndsize,sizeof(sndsize)) == SOCKET_ERROR) {
              break;
        }
    }

    length=sizeof(sndsize);
    if (getsockopt(tsSocket,SOL_SOCKET,SO_SNDBUF,(char *)&sndsize,&length) == SOCKET_ERROR) {
        ReportError();
        sprintf(MsgBuf, "getsockopt() failed:  ");
        fprintf (stdout, "%s\n", MsgBuf);
    }
    
    sprintf(MsgBuf, "TCP socket send buffer was %d - now %d",temp,sndsize);
    if (verbose) fprintf (stdout, "%s\n", MsgBuf);

/*    repeat for receive buffer which seems to also be needed     */

    length=sizeof(rcvsize);
    if (getsockopt(tsSocket,SOL_SOCKET,SO_RCVBUF,(char *)&rcvsize,&length) == SOCKET_ERROR) {
        ReportError();
        sprintf(MsgBuf, "getsockopt() failed:  ");
        fprintf (stdout, "%s\n", MsgBuf);
    }
    temp = rcvsize;

    for (rcvreq = MINRCVBUFSIZE; rcvreq <= RcvBufSize; rcvreq = rcvreq+MINRCVBUFSIZE) {
        rcvsize = rcvreq*1024;
        if (setsockopt(tsSocket,SOL_SOCKET,SO_RCVBUF,(char *)&rcvsize,sizeof(rcvsize)) == SOCKET_ERROR) {
              break;
        }
    }

    length=sizeof(rcvsize);
    if (getsockopt(tsSocket,SOL_SOCKET,SO_RCVBUF,(char *)&rcvsize,&length) == SOCKET_ERROR) {
        ReportError();
        sprintf(MsgBuf, "getsockopt() failed:  ");
        fprintf (stdout, "%s\n", MsgBuf);
    }
    
    sprintf(MsgBuf, "TCP socket receive buffer was %d - now %d",temp,rcvsize);
    if (verbose) fprintf (stdout, "%s\n", MsgBuf);

    sprintf(MsgBuf, "TCP socket created OK - now connecting to %s",IPStr);
    if (verbose) fprintf (stdout, "%s\n", MsgBuf);

    if (connect(tsSocket,(struct sockaddr*)&server,sizeof(server)) == SOCKET_ERROR) {
        ReportError();
        sprintf(MsgBuf, "connect() failed:  ");
        fprintf (stdout, "%s\n", MsgBuf);
        XportClose();
        return XPORT_ERROR;
    }

    //    set the socket to non-blocking mode

#ifdef WIN32
    if (ioctlsocket(tsSocket, FIONBIO, (unsigned long *)&yes) == SOCKET_ERROR) {
        ReportError();
        sprintf(MsgBuf,"ioctlsocket() failed: ");
        fprintf (stdout, "%s\n", MsgBuf);
        XportClose();
        return XPORT_ERROR;
    }
#endif
#ifdef UNIX
    if (fcntl(tsSocket, F_SETFL, O_NONBLOCK) == -1) {   /* set socket in non blocking mode */
        ReportError();
        sprintf(MsgBuf, "fcntl() failed:  ");
        fprintf (stdout, "%s\n", MsgBuf);
        XportClose();
        return XPORT_ERROR;
    }
#endif

    sprintf(MsgBuf, "Connected to %s port %d",IPStr, Port);
    if (verbose) fprintf (stdout, "%s\n", MsgBuf);

    return OK;
}
/*******************************************************************************
* XportClose            final function to close host/client connection
*                        destroys a socket
*/
void XportClose ()
{

    if (tsSocket != 0) {
#ifdef WIN32
             closesocket (tsSocket);
             WSACleanup();
#else
             close (tsSocket);
#endif
             tsSocket = 0;
    }
    
}
/*******************************************************************************
* clientInit   initialise 
*              initialise control structures
*              initialise client structure
*/
int clientInit (char *host)
{
    int rs;

    XportClose ();
    strcpy (XportHost, host);
    rs = XportOpenClient (XportHost);
    return (rs);
}
/********************************************************************************
* XportTxData       Transmit the data 
*/
int XportTxData (unsigned char *data, int *len)
{
        int retval;
        int length;
        int cmd;

/*    if no connection currently exists - open one */

    if (tsSocket == 0)  (void)clientInit (XportHost);
    if (tsSocket == 0)  return XPORT_ERROR;

       length = (*len)-4;
//       memcpy((char*)data+1, (char*) &length+1, 3);    /* ensure message length is correct */

       cmd = ((*data) >> 5) & 3;   /* write=0; long write=1; read=2 */

/*    send the data block  */

        length = *len;

        retval = do_send(tsSocket,(char *)data,length);
        if (retval == SOCKET_ERROR) {
            ReportError();
            sprintf(MsgBuf, "send() failed: ");
            fprintf (stdout, "%s\n", MsgBuf);
            XportClose();
            return XPORT_ERROR;
        }
        if (retval == TIMEOUT) {
            sprintf(MsgBuf, "connection timeout in send()");
            fprintf (stdout, "%s\n", MsgBuf);
            XportClose();
            return XPORT_ERROR;
        }

        if (verbose) printf("Sent %d bytes to server\n",length);

/*    wait for acknowledgement  */

        length = sizeof (HEADER);
        retval = do_recv(tsSocket,(char *)data,&length,0 );
        if (retval == SOCKET_ERROR) {
            ReportError();
            sprintf(MsgBuf, "recv() failed: ");
            fprintf (stdout, "%s\n", MsgBuf);
            printf("Received %d bytes from server\n",length);
            XportClose();
            return XPORT_ERROR;
        }
        if (retval == TIMEOUT) {
            sprintf(MsgBuf, "connection timeout in recv() #1");
            fprintf (stdout, "%s\n", MsgBuf);
            printf("Received %d bytes from server\n",length);
            XportClose();
            return XPORT_ERROR;
        }
        if (retval != OK) {
            sprintf(MsgBuf, "Server closed connection\n");
            fprintf (stdout, "%s\n", MsgBuf);
            XportClose();
            return XPORT_ERROR;
        }
        if (verbose) printf("Received %d bytes from server\n",length);
                      
        length = (((data[1] << 8) + data[2]) << 8) + data[3];
        if (verbose) printf("require %d bytes more from server\n",length);
        if (length < 0) {
            printf("server response is %d bytes (0x%x 0x%x 0x%x) which is < 0\n", length, data[1], data[2], data[3]);
            printf("this is not acceptable\n");
            XportClose();
            return XPORT_ERROR;
        }
        if (length > BlockSize) {
            printf("server response is %d bytes (0x%x 0x%x 0x%x) which is greater than %d bytes\n", length, data[1], data[2], data[3], BlockSize);
            printf("this is not acceptable\n");
            XportClose();
            return XPORT_ERROR;
        }
        if (length > 0) {
            retval = do_recv(tsSocket,(char *)&data[4],&length,0 );
            if (retval == SOCKET_ERROR) {
                ReportError();
                sprintf(MsgBuf, "recv() failed: ");
                fprintf (stdout, "%s\n", MsgBuf);
                printf("Received %d bytes from server\n",length);
                XportClose();
                return XPORT_ERROR;
            }
            if (retval == TIMEOUT) {
                sprintf(MsgBuf, "connection timeout in recv() #2");
                fprintf (stdout, "%s\n", MsgBuf);
                printf("Received %d bytes from server\n",length);
                XportClose();
                return XPORT_ERROR;
            }
            if (retval != OK) {
                sprintf(MsgBuf, "Server closed connection\n");
                fprintf (stdout, "%s\n", MsgBuf);
                XportClose();
                return XPORT_ERROR;
            }
            if (verbose) printf("Received %d bytes from server\n",length);
        }

    *len = length+4;
    return OK;
}
/*******************************************************************************/
